# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE SETUP.PY
short_version = '4.22.0'
version = '4.22.0'
full_version = '4.22.0'
git_revision = 'afa06382ea8e6635cad86d3f6ead16bdc8548963'
release = True

if not release:
    version = full_version
    short_version += ".dev"
